#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
Except: snippet of code from Mark@StackOverflow
"""

import base64
import json

####################################################################################################
# The following code snippet comes from:
# Author: Mark (https://stackoverflow.com/users/2606953/mark)
# Source: https://stackoverflow.com/questions/40000495/
#                                   how-to-encode-bytes-in-json-json-dumps-throwing-a-typeerror
class BytesEncoder(json.JSONEncoder):
    """ Custom class to convert binary fields to strings in JSON """
    def default(self, o):
        if isinstance(o, bytes):
            return base64.b64encode(o).decode("ascii")
        return super().default(o)

####################################################################################################
def oab2json(data, compact=False, indent=4):
    """ Convert our dictionary results to JSON """
    results = None

    if compact:
        results = json.dumps(data, cls=BytesEncoder)
    else:
        results = json.dumps(data, indent=indent, cls=BytesEncoder)

    return results



#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

####################################################################################################
def oab2csv(data, d):
    """ Print our dictionary results in CSV format using the given delimiter to separate values """
    users = []
    users.append(
        f'"Name"{d}"Title"{d}"Surname"{d}"Given name"{d}"Initials"' +
        f'{d}"Company"{d}"Account"{d}"Email"{d}"Mobile"' +
        f'{d}"Business phone"{d}"Pager"{d}"Fax"{d}"Home phone"' +
        f'{d}"Department"{d}"Office location"{d}"Street address"' +
        f'{d}"Postal code"{d}"Locality"{d}"State or province"{d}"Country"'
    )

    lists = []
    lists.append(
        f'"Name"{d}"Account"{d}"Email"{d}"Total members"{d}"Internal members"' +
        f'{d}"External members"{d}"Comment"'
    )

    count_users = 0
    count_lists = 0
    count_lists_without_members = 0
    count_unknown = 0

    for entry in data["address_book"]:
        record = entry["record"]

        # Sanitize string values
        for key, value in record.items():
            if isinstance(value, str):
                record[key] = value.replace('"', "'").replace("\r", ". ")

        match record["PidTagObjectType"]:
            case 6: # User
                count_users += 1
                display_name = ""
                if "PidTagDisplayName" in record:
                    display_name = record["PidTagDisplayName"]
                title = ""
                if "PidTagTitle" in record:
                    title = record["PidTagTitle"]
                surname = ""
                if "PidTagSurname" in record:
                    surname = record["PidTagSurname"]
                given_name = ""
                if "PidTagGivenName" in record:
                    given_name = record["PidTagGivenName"]
                initials = ""
                if "PidTagInitials" in record:
                    initials = record["PidTagInitials"]
                company_name = ""
                if "PidTagCompanyName" in record:
                    company_name = record["PidTagCompanyName"]

                account = ""
                if "PidTagAddressBookDisplayNamePrintable" in record:
                    account = record["PidTagAddressBookDisplayNamePrintable"]
                elif "PidTagAccount" in record:
                    account = record["PidTagAccount"]
                address = ""
                if "PidTagSmtpAddress" in record:
                    address = record["PidTagSmtpAddress"]
                mobile = ""
                if "PidTagMobileTelephoneNumber" in record:
                    mobile = record["PidTagMobileTelephoneNumber"]
                business_phone = ""
                if "PidTagBusinessTelephoneNumber" in record:
                    business_phone = record["PidTagBusinessTelephoneNumber"]
                if "PidTagBusiness2TelephoneNumbers" in record:
                    for number in record["PidTagBusiness2TelephoneNumbers"]:
                        business_phone += ',' + number.replace('"', "'")
                pager = ""
                if "PidTagPagerTelephoneNumber" in record:
                    pager = record["PidTagPagerTelephoneNumber"]
                fax = ""
                if "PidTagPrimaryFaxNumber" in record:
                    fax = record["PidTagPrimaryFaxNumber"]
                home_phone = ""
                if "PidTagHomeTelephoneNumber" in record:
                    home_phone = record["PidTagHomeTelephoneNumber"]
                if "PidTagHome2TelephoneNumbers" in record:
                    for number in record["PidTagHome2TelephoneNumbers"]:
                        home_phone += ',' + number.replace('"', "'")

                department_name = ""
                if "PidTagDepartmentName" in record:
                    department_name = record["PidTagDepartmentName"]
                office_location = ""
                if "PidTagOfficeLocation" in record:
                    office_location = record["PidTagOfficeLocation"]
                street_address = ""
                if "PidTagStreetAddress" in record:
                    street_address = record["PidTagStreetAddress"]
                postal_code = ""
                if "PidTagPostalCode" in record:
                    postal_code = record["PidTagPostalCode"]
                locality = ""
                if "PidTagLocality" in record:
                    locality = record["PidTagLocality"]
                state = ""
                if "PidTagStateOrProvince" in record:
                    state = record["PidTagStateOrProvince"]
                country = ""
                if "PidTagCountry" in record:
                    country = record["PidTagCountry"]

                users.append(
                    f'"{display_name}"{d}"{title}"{d}"{surname}"{d}"{given_name}"{d}"{initials}"' +
                    f'{d}"{company_name}"{d}"{account}"{d}"{address}"{d}"{mobile}"' +
                    f'{d}"{business_phone}"{d}"{pager}"{d}"{fax}"{d}"{home_phone}"' +
                    f'{d}"{department_name}"{d}"{office_location}"{d}"{street_address}"' +
                    f'{d}"{postal_code}"{d}"{locality}"{d}"{state}"{d}"{country}"'
                )

            case 8: # List
                count_lists += 1
                display_name = ""
                if "PidTagDisplayName" in record:
                    display_name = record["PidTagDisplayName"]
                account = ""
                if "PidTagAddressBookDisplayNamePrintable" in record:
                    account = record["PidTagAddressBookDisplayNamePrintable"]
                elif "PidTagAccount" in record:
                    account = record["PidTagAccount"]
                address = ""
                if "PidTagSmtpAddress" in record:
                    address = record["PidTagSmtpAddress"]
                members = 0
                internal_members = 0
                external_members = 0
                if "PidTagAddressBookDistributionListMemberCount" in record:
                    internal_members = record["PidTagAddressBookDistributionListMemberCount"]
                if "PidTagAddressBookDistributionListExternalMemberCount" in record:
                    external_members = \
                        record["PidTagAddressBookDistributionListExternalMemberCount"]
                members = internal_members + external_members
                if not members:
                    count_lists_without_members += 1
                comment = ""
                if "PidTagComment" in record:
                    comment = record["PidTagComment"]

                lists.append(
                    f'"{display_name}"{d}"{account}"{d}"{address}"{d}{members}' +
                    f'{d}{internal_members}{d}{external_members}{d}"{comment}"'
                )

            case _: # Unknown
                count_unknown += 1

    return users, lists



#!/usr/bin/env python3
""" oab - Outlook Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
"""

import binascii
import io
import logging
import math
import os
import re
import struct

import rich.progress

from .properties import property_tags

####################################################################################################
def _get_file_type(file_name):
    """ Returns an OAB file type from its name """
    # The name of change files or compressed files is not known
    match file_name:
        case "ubrowse.oab":
            file_type = "browse file"
        case "uANRdex.oab":
            file_type = "ANR index file"
        case "udetails.oab":
            file_type = "details file"
        case "uPDNdex.oab":
            file_type = "PDN index file"
        case "uRDNdex.oab":
            file_type = "RDN index file"
        case "utmplts.oab":
            file_type = "display template file"
        case _:
            file_type = "unknown file"

    return file_type

####################################################################################################
def _read_OAB_HDR(file_type, file):
    """ Returns the OAB file header (OAB_HDR) """
    data = {}

    # Reading 3 little-endian unsigned int
    (data["ulVersion"], data["ulSerial"], data["ulTotRecs"]) = \
        struct.unpack('<III', file.read(4 * 3))
    match file_type:
        case "browse file":
            match data["ulVersion"]:
                case 0xa:
                    data["version"] = "uncompressed version 2 OAB browse file"
                case 0xe:
                    data["version"] = "uncompressed version 3 OAB browse file"
                case _:
                    data["version"] = "unknown file"

        case "ANR index file":
            match data["ulVersion"]:
                case 0xa:
                    data["version"] = "uncompressed version 2 ANR index file"
                case 0xe:
                    data["version"] = "uncompressed version 3 ANR index file"
                case _:
                    data["version"] = "unknown file"

        case "PDN index file":
            match data["ulVersion"]:
                case _:
                    data["version"] = "unknown file"

        case "RDN index file":
            data["oRoot"] = struct.unpack('<I', file.read(4))[0]
            match data["ulVersion"]:
                case 0xa:
                    data["version"] = "uncompressed version 2 RDN index file"
                case 0xe:
                    data["version"] = "uncompressed version 3 RDN index file"
                case _:
                    data["version"] = "unknown file"

        case "display template file":
            match data["ulVersion"]:
                case 0x7:
                    data["version"] = "uncompressed display template file"
                case _:
                    data["version"] = "unknown file"

        case "change file":
            match data["ulVersion"]:
                case 0xb:
                    data["version"] = "uncompressed version 2 changes file"
                case 0xf:
                    data["version"] = "uncompressed version 3 changes file"
                case _:
                    data["version"] = "unknown file"

        case "compressed file":
            data["ulVersionHi"] = data["ulVersion"]
            del data["ulVersion"]
            data["ulVersionLo"] = data["ulSerial"]
            del data["ulSerial"]
            data["ulBlockMax"] = data["ulTotRecs"]
            del data["ulTotRecs"]
            data["ulTargetSize"] = struct.unpack('<I', file.read(4))[0]
            if data["ulVersionHi"] == 0x2 and data["ulVersionLo"] == 0x1:
                data["version"] = "compressed version 2 or 3 OAB file"
            elif data["ulVersionHi"] == 0x3 and data["ulVersionLo"] == 0x1:
                data["version"] = "compressed version 4 OAB file"
            else:
                data["version"] = "unknown file"

        case _: # including "details file"
            match data["ulVersion"]:
                case 0x7:
                    data["version"] = "uncompressed version 2 or 3 details file"
                case 0x20:
                    data["version"] = "uncompressed version 4 full details file"
                case _:
                    data["version"] = "unknown file"

    return data

####################################################################################################
def _read_OAB_PROP_TABLE(bdata):
    """ Returns an OAB_PROP_TABLE structure contents """
    data = {}
    data["cAtts"] = struct.unpack('<I', bdata.read(4))[0]
    data["rgProps"] = []
    for _ in range(data["cAtts"]):
        ulPropId = struct.unpack('<I', bdata.read(4))[0]
        ulFlags  = struct.unpack('<I', bdata.read(4))[0]

        usPropType = ulPropId & 0x0000FFFF
        match usPropType:
            case 0x0003: property_type = "PtypInteger32"
            case 0x000B: property_type = "PtypBoolean"
            case 0x000D: property_type = "PtypObject"
            case 0x001E: property_type = "PtypString8"
            case 0x001F: property_type = "PtypString"
            case 0x0102: property_type = "PtypBinary"
            case 0x1003: property_type = "PtypMultipleInteger32"
            case 0x101E: property_type = "PtypMultipleString8"
            case 0x101F: property_type = "PtypMultipleString"
            case 0x1102: property_type = "PtypMultipleBinary"
            case _: property_type = "unknown"

        if ulPropId in property_tags:
            data["rgProps"].append((property_tags[ulPropId], property_type, ulFlags))
        else:
            data["rgProps"].append((str(ulPropId), property_type, ulFlags))

    return data

####################################################################################################
def _read_OAB_META_DATA(file):
    """ Returns the OAB metadata record (OAB_META_DATA) """
    data = {}
    data["cbSize"] = struct.unpack('<I', file.read(4))[0]

    meta_data = io.BytesIO(file.read(data["cbSize"] - 4))
    data["rgHdrAtts"] = _read_OAB_PROP_TABLE(meta_data)
    data["rgOabAtts"] = _read_OAB_PROP_TABLE(meta_data)

    return data

####################################################################################################
def _read_PtypInteger32(bdata):
    """ Returns an integer from a PTypInteger32 """
    byte_count = struct.unpack('<B', bdata.read(1))[0]
    if 0x81 <= byte_count <= 0x84:
        byte_count = struct.unpack('<I', (bdata.read(byte_count - 0x80) + b"\0\0\0")[0:4])[0]
    elif byte_count > 127:
        return -1

    return byte_count

####################################################################################################
def _read_PtypBoolean(bdata):
    """ Returns a boolean from a PTypBoolean """
    return struct.unpack('<?', bdata.read(1))[0]

####################################################################################################
def _read_PtypString(bdata):
    """ Returns a string from a PTypString or PTypString8 """
    buffer = b""
    while True:
        character = bdata.read(1)
        if character in (b"\0", b""):
            break
        buffer += character

    return buffer.decode('utf-8', errors="ignore")

####################################################################################################
def _read_PtypBinary(bdata):
    """ Returns an hexadecimal string from a PTypBinary """
    length = _read_PtypInteger32(bdata)
    binary = bdata.read(length)

    return binascii.b2a_hex(binary)

####################################################################################################
def _read_PtypMultipleString(bdata):
    """ Returns a list of strings from a PTypMultipleString or PTypMultipleString8 """
    byte_count = _read_PtypInteger32(bdata)
    strings_list = []
    for _ in range(byte_count):
        strings_list.append(_read_PtypString(bdata))

    return strings_list

####################################################################################################
def _read_PtypMultipleInteger32(bdata):
    """ Returns a list of integers from a PTypMultipleInteger32 """
    byte_count = _read_PtypInteger32(bdata)
    int_list = []
    for _ in range(byte_count):
        int_list.append(_read_PtypInteger32(bdata))

    return int_list

####################################################################################################
def _read_PtypMultipleBinary(bdata):
    """ Returns a list of hexadecimal strings from a PTypMultipleBinary """
    byte_count = _read_PtypInteger32(bdata)
    bin_list = []
    for _ in range(byte_count):
        bin_list.append(_read_PtypBinary(bdata))

    return bin_list

####################################################################################################
def _read_OAB_V4_REC(file, atts_count, attributes):
    """ Returns the OAB_V4_RECrecord """
    data = {}
    data["cbSize"] = struct.unpack('<I', file.read(4))[0]

    bdata = io.BytesIO(file.read(data["cbSize"] - 4))
    presenceBitArray = bytearray(bdata.read(int(math.ceil(atts_count / 8.0))))
    indices = [i for i in range(atts_count) if (presenceBitArray[i // 8] >> (7 - (i % 8))) & 1 == 1]

    data["record"] = {}
    for i in indices:
        attribute_name = attributes[i][0]
        attribute_type = attributes[i][1]

        match attribute_type:
            case "PtypInteger32":
                data["record"][attribute_name] = _read_PtypInteger32(bdata)
            case "PtypBoolean":
                data["record"][attribute_name] = _read_PtypBoolean(bdata)
            case "PtypObject":
                raise ValueError("Unhandled PTypObject attribute type")
            case "PtypString8":
                data["record"][attribute_name] = _read_PtypString(bdata)
            case "PtypString":
                data["record"][attribute_name] = _read_PtypString(bdata)
            case "PtypBinary":
                data["record"][attribute_name] = _read_PtypBinary(bdata)
            case "PtypMultipleInteger32":
                values = _read_PtypMultipleInteger32(bdata)
                if attribute_name == "PidTagOfflineAddressBookTruncatedProperties":
                    new_values = []
                    for value in values:
                        if value in property_tags:
                            new_values.append(property_tags[value])
                        else:
                            new_values.append(str(value))
                    values = new_values
                data["record"][attribute_name] = values
            case "PtypMultipleString8":
                data["record"][attribute_name] = _read_PtypMultipleString(bdata)
            case "PtypMultipleString":
                data["record"][attribute_name] = _read_PtypMultipleString(bdata)
            case "PtypMultipleBinary":
                data["record"][attribute_name] = _read_PtypMultipleBinary(bdata)
            case _:
                raise ValueError(f"Unhandled {attribute_type} attribute type")

    return data

####################################################################################################
def load_oab_file(path_name, progress_bar=False):
    """ Returns a dictionary with the OAB file contents """
    data = {"path_name": path_name}
    data["name"] = os.path.basename(path_name)

    # Determining file type
    data["type"] = _get_file_type(data["name"])
    if data["type"] == "unknown file":
        logging.warning("Unknown file type. Assuming an OAB details file")

    with open(path_name, 'rb') as file:
        data["OAB_HDR"] = _read_OAB_HDR(data["type"], file)

        if data["OAB_HDR"]["version"] == "uncompressed version 4 full details file":
            data["OAB_META_DATA"] = _read_OAB_META_DATA(file)
            data["header_record"] = _read_OAB_V4_REC(
                file,
                data["OAB_META_DATA"]["rgHdrAtts"]["cAtts"],
                data["OAB_META_DATA"]["rgHdrAtts"]["rgProps"]
            )

            if progress_bar:
                to_be_processed = rich.progress.track(
                    range(data["OAB_HDR"]["ulTotRecs"]),
                    description="Processing"
                )
            else:
                to_be_processed = range(data["OAB_HDR"]["ulTotRecs"])

            data["address_book"] = []
            for _ in to_be_processed:
                data["address_book"].append(
                    _read_OAB_V4_REC(
                        file,
                        data["OAB_META_DATA"]["rgOabAtts"]["cAtts"],
                        data["OAB_META_DATA"]["rgOabAtts"]["rgProps"]
                    )
                )
        else:
            logging.error("%s are not supported", data["OAB_HDR"]["version"])

    return data



#!/usr/bin/env python3
""" oab - Offline Address Books decoder
License: 3-clause BSD (see https://opensource.org/licenses/BSD-3-Clause)
Author: Hubert Tournier
Except: snippet of code from Wolph@StackOverflow
"""

import contextlib
import getopt
import logging
import os
import pprint
import sys

import libpnu

from .library import load_oab_file
from .csv_output import oab2csv
from .excel_output import oab2excel
from .json_output import oab2json
from .table_output import oab2table

# Version string used by the what(1) and ident(1) commands:
ID = "@(#) $Id: oab - Offline Address Books decoder v1.0.1 (June 22, 2024) by Hubert Tournier $"

# Default parameters. Can be overcome by environment variables, then command line options
parameters = {
    "format": "txt",
    "progress_bar": True,
    "filename": "",
    # Parameters for CSV format
    "delimiter": ',',
    # Parameters for JSON format
    "compact": False,
    "indent": 4,
    # Parameters for PrettyTable format
    "width": None,
}

####################################################################################################
def _display_help():
    """ Display usage and help """
    #pylint: disable=C0301
    print("usage: oab [--debug] [--help|-?] [--version]", file=sys.stderr)
    print("       [--bar|-b] [--output|-o FILE]", file=sys.stderr)
    print("       [--csv|-C] [--delimiter|-d CHAR] [--excel|-E]", file=sys.stderr)
    print("       [--json|-J] [--compact|-c] [--ident|-i INT]", file=sys.stderr)
    print("       [--table|-T] [--width|-w INT]", file=sys.stderr)
    print("       [--] filename", file=sys.stderr)
    print("  -------------------  --------------------------------------------------", file=sys.stderr)
    print("  --bar|-b             Toggle OFF progress bar", file=sys.stderr)
    print("  --output|-o FILE     Output result into FILE (extension added by oab)", file=sys.stderr)
    print("  -------------------  --------------------------------------------------", file=sys.stderr)
    print("  --csv|-C             Format results as CSV", file=sys.stderr)
    print("  --delimiter|-d CHAR  Use CHAR as delimiter for CSV format", file=sys.stderr)
    print("  -------------------  --------------------------------------------------", file=sys.stderr)
    print("  --excel|-E           Format results as Excel tabbed file", file=sys.stderr)
    print("  -------------------  --------------------------------------------------", file=sys.stderr)
    print("  --json|-J            Format results as JSON", file=sys.stderr)
    print("  --compact|-c         Use compact JSON format", file=sys.stderr)
    print("  --ident|-i INT       Use INT spaces for indented JSON format", file=sys.stderr)
    print("  -------------------  --------------------------------------------------", file=sys.stderr)
    print("  --table|-T           Format results as tables", file=sys.stderr)
    print("  --width|-w INT       Use INT columns in table format", file=sys.stderr)
    print("  -------------------  --------------------------------------------------", file=sys.stderr)
    print("  --debug              Enable debug mode", file=sys.stderr)
    print("  --help|-?            Print usage and this help message and exit", file=sys.stderr)
    print("  --version            Print version and exit", file=sys.stderr)
    print("  --                   Options processing terminator", file=sys.stderr)
    print(file=sys.stderr)
    #pylint: enable=C0301

####################################################################################################
def _process_command_line():
    """ Process command line options """
    #pylint: disable=C0103, W0602
    global parameters
    #pylint: enable=C0103, W0602

    # option letters followed by : expect an argument
    # same for option strings followed by =
    character_options = "CEJTbcd:i:o:w:?"
    string_options = [
        "bar",
        "compact",
        "csv",
        "debug",
        "delimiter=",
        "excel",
        "help",
        "indent=",
        "json",
        "output=",
        "table",
        "version",
        "width=",
    ]

    try:
        options, remaining_arguments = getopt.getopt(
            sys.argv[1:], character_options, string_options
        )
    except getopt.GetoptError as error:
        logging.critical("Syntax error: %s", error)
        _display_help()
        sys.exit(1)

    for option, argument in options:

        if option == "--debug":
            logging.disable(logging.NOTSET)

        elif option in ("--help", "-?"):
            _display_help()
            sys.exit(0)

        elif option == "--version":
            print(ID.replace("@(" + "#)" + " $" + "Id" + ": ", "").replace(" $", ""))
            sys.exit(0)

        elif option in ("--bar", "-b"):
            parameters["progress_bar"] = False

        elif option in ("--compact", "-c"):
            parameters["compact"] = True

        elif option in ("--csv", "-C"):
            parameters["format"] = "csv"

        elif option in ("--delimiter", "-d"):
            parameters["delimiter"] = argument[0]

        elif option in ("--excel", "-E"):
            parameters["format"] = "excel"

        elif option in ("--indent", "-i"):
            try:
                parameters["indent"] = int(argument)
            except ValueError:
                logging.warning("Invalid argument for --indent option. Keeping default value")
            if parameters["indent"] < 0:
                logging.warning("Invalid argument for --indent option. Keeping default value")
                parameters["indent"] = 4

        elif option in ("--json", "-J"):
            parameters["format"] = "json"

        elif option in ("--output", "-o"):
            parameters["filename"] = argument

        elif option in ("--table", "-T"):
            parameters["format"] = "table"

        elif option in ("--width", "-w"):
            try:
                parameters["width"] = int(argument)
            except ValueError:
                logging.warning("Invalid argument for --width option. Keeping default value")
            if parameters["width"] < 80:
                logging.warning("Invalid argument for --width option. Keeping default value")
                parameters["width"] = None

    return remaining_arguments

####################################################################################################
# The following code is from:
# Author: Wolph https://stackoverflow.com/users/54017/wolph
# Source: https://stackoverflow.com/questions/17602878/
#                                   how-to-handle-both-with-open-and-sys-stdout-nicely
@contextlib.contextmanager
def smart_open(filename=None):
    """ Opens files and stdout streams in the same way """
    if filename and filename != '-':
        fh = open(filename, 'w', encoding="utf-8", errors="ignore")
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()

####################################################################################################
def main():
    """ The program's main entry point """
    program_name = os.path.basename(sys.argv[0])

    libpnu.initialize_debugging(program_name)
    libpnu.handle_interrupt_signals(libpnu.interrupt_handler_function)
    arguments = _process_command_line()

    if not arguments:
        _display_help()
    else:
        argument = arguments[0]
        if os.path.isfile(argument):
            data = load_oab_file(argument, progress_bar=parameters["progress_bar"])
            match parameters["format"]:
                case "csv":
                    users, lists = oab2csv(data, parameters["delimiter"])
                    filename = parameters["filename"]
                    if filename:
                        filename += ".users.csv"
                        print(f"Writing to {filename}")
                    with smart_open(filename) as file:
                        for item in users:
                            print(item, file=file)

                    filename = parameters["filename"]
                    if filename:
                        filename += ".lists.csv"
                        print(f"Writing to {filename}")
                    else:
                        print()
                    with smart_open(filename) as file:
                        for item in lists:
                            print(item, file=file)

                case "excel":
                    filename = parameters["filename"]
                    if filename:
                        filename += ".xlsx"
                    else:
                        filename = "address_book.xlsx"
                    print(f"Writing to {filename}")
                    oab2excel(data, filename)

                case "json":
                    filename = parameters["filename"]
                    if filename:
                        filename += ".json"
                        print(f"Writing to {filename}")
                    with smart_open(filename) as file:
                        print(
                            oab2json(
                                data,
                                compact=parameters["compact"],
                                indent=parameters["indent"],
                            ),
                            file=file
                        )

                case "table":
                    users, lists = oab2table(data, width=parameters["width"])
                    filename = parameters["filename"]
                    if filename:
                        filename += ".users.txt"
                        print(f"Writing to {filename}")
                    with smart_open(filename) as file:
                        print(users, file=file)

                    filename = parameters["filename"]
                    if filename:
                        filename += ".lists.txt"
                        print(f"Writing to {filename}")
                    else:
                        print()
                    with smart_open(filename) as file:
                        print(lists, file=file)

                case "txt":
                    filename = parameters["filename"]
                    if filename:
                        filename += ".txt"
                        print(f"Writing to {filename}")
                    with smart_open(filename) as file:
                        pprint.pprint(data, file)

                case "xml":
                    # Who cares ?
                    pass
        else:
            logging.error("'%s' is not a readable file", argument)
            sys.exit(1)

    sys.exit(0)

if __name__ == "__main__":
    main()
